"""
JSON5 parser using regex and native JSON parsing.
Compatible with PyPy and doesn't require external dependencies.

This parser handles:
- Single-line comments (//)
- Multi-line comments (/* */)
- Unquoted object keys
- Trailing commas
- Single-quoted strings
"""

import json
import logging
import re
from typing import Any

logger = logging.getLogger(__name__)


def parse_json5(content: str, debug: bool = False) -> dict[str, Any]:
    """
    Parse JSON5 content and return a dictionary.

    Args:
        content: JSON5 string content
        debug: If True, print debug output on parse errors (default: False)

    Returns:
        Dictionary representation of the JSON5 content

    Raises:
        json.JSONDecodeError: If the content cannot be parsed
    """
    # Process the content in a single pass to:
    # 1. Remove comments (but not // or /* */ inside strings)
    # 2. Convert single-quoted strings to double-quoted strings
    # 3. Preserve all string content correctly

    def process_content(text):
        """Process JSON5 content: remove comments and convert quotes."""
        result = []
        i = 0

        while i < len(text):
            char = text[i]

            # Check for double-quoted strings - preserve as-is
            if char == '"':
                # Copy the entire double-quoted string, handling escapes
                result.append(char)
                i += 1
                while i < len(text):
                    char = text[i]
                    result.append(char)
                    if char == "\\" and i + 1 < len(text):
                        # Copy the escaped character
                        i += 1
                        result.append(text[i])
                        i += 1
                    elif char == '"':
                        # End of string
                        i += 1
                        break
                    else:
                        i += 1
                continue

            # Check for single-quoted strings - convert to double quotes
            if char == "'":
                # Start of single-quoted string
                result.append('"')  # Convert to double quote
                i += 1
                while i < len(text):
                    char = text[i]
                    if char == "\\" and i + 1 < len(text):
                        # Handle escape sequences
                        next_char = text[i + 1]
                        if next_char == "'":
                            # Escaped single quote in single-quoted string
                            # In double-quoted string, we don't need to escape it
                            result.append("'")
                            i += 2
                        else:
                            # Other escape sequences - preserve
                            result.append(char)
                            i += 1
                            result.append(text[i])
                            i += 1
                    elif char == "'":
                        # End of single-quoted string
                        result.append('"')  # Convert to double quote
                        i += 1
                        break
                    elif char == '"':
                        # Double quote inside single-quoted string - escape it
                        result.append('\\"')
                        i += 1
                    else:
                        result.append(char)
                        i += 1
                continue

            # Check for multi-line comment start
            if i + 1 < len(text) and text[i : i + 2] == "/*":  # noqa: E203
                # Find the end of the comment
                end = text.find("*/", i + 2)
                if end != -1:
                    # Skip the entire comment, but preserve newlines for line number accuracy
                    comment_text = text[i : end + 2]  # noqa: E203
                    newlines = comment_text.count("\n")
                    result.append("\n" * newlines)
                    i = end + 2
                else:
                    # Unclosed comment, skip to end
                    i = len(text)
                continue

            # Check for single-line comment
            # Make sure it's not part of a URL (http:// or https://)
            if i + 1 < len(text) and text[i : i + 2] == "//":  # noqa: E203
                # Check if this is part of a URL
                if i > 0 and text[i - 1] == ":":
                    # This is likely a URL, keep it
                    result.append(char)
                    i += 1
                else:
                    # This is a comment - skip to end of line
                    while i < len(text) and text[i] != "\n":
                        i += 1
                    # Keep the newline
                    if i < len(text):
                        result.append("\n")
                        i += 1
                continue

            # Regular character
            result.append(char)
            i += 1

        return "".join(result)

    content = process_content(content)

    # Convert unquoted keys to quoted keys
    # Match: word characters followed by colon (not already quoted)
    # This regex looks for keys at the start of a line or after { or ,
    content = re.sub(r"([\{\,]\s*)([a-zA-Z_$][a-zA-Z0-9_$]*)\s*:", r'\1"\2":', content)

    # Also handle keys at the beginning of the content
    content = re.sub(
        r"^(\s*)([a-zA-Z_$][a-zA-Z0-9_$]*)\s*:",
        r'\1"\2":',
        content,
        flags=re.MULTILINE,
    )

    # Remove trailing commas before closing braces/brackets
    content = re.sub(r",(\s*[}\]])", r"\1", content)

    # Parse as regular JSON
    try:
        return json.loads(content)
    except json.JSONDecodeError as e:
        # Log debug information if requested
        if debug:
            logger.error("\n" + "=" * 80)
            logger.error("JSON5 Parser Error - Processed content that failed to parse:")
            logger.error("=" * 80)
            lines = content.split("\n")
            for i, line in enumerate(lines, 1):
                marker = " --> " if i == e.lineno else "     "
                logger.error(f"{marker}{i:3}: {line}")
            logger.error("=" * 80)
            logger.error(f"Error at line {e.lineno}, column {e.colno}: {e.msg}")
            logger.error("=" * 80 + "\n")
        raise


def encode_json5(data: dict[str, Any], indent: int = 2) -> str:
    """
    Encode dictionary to JSON5 format (actually just JSON with nice formatting).

    Args:
        data: Dictionary to encode
        indent: Number of spaces for indentation

    Returns:
        JSON5-formatted string
    """
    return json.dumps(data, indent=indent, ensure_ascii=False)


def load_json5(file_path: str, debug: bool = False) -> dict[str, Any]:
    """
    Load and parse a JSON5 file.

    Args:
        file_path: Path to JSON5 file
        debug: If True, print debug output on parse errors (default: False)

    Returns:
        Dictionary representation of the JSON5 content
    """
    with open(file_path, encoding="utf-8") as f:
        content = f.read()
    return parse_json5(content, debug=debug)


def dump_json5(data: dict[str, Any], file_path: str, indent: int = 2):
    """
    Dump dictionary to a JSON5 file (as regular JSON).

    Args:
        data: Dictionary to dump
        file_path: Path to output file
        indent: Number of spaces for indentation
    """
    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=indent, ensure_ascii=False)



"""
Comprehensive test suite for the JSON5 parser.

Tests cover:
- Comment removal (single-line and multi-line)
- Quote conversion (single to double)
- Unquoted keys
- Trailing commas
- Edge cases with strings containing special characters
- URL preservation in strings
- Mixed quote types
- Escape sequences
- Complex realistic scenarios
"""

import json
import os
import tempfile
import unittest

from npm_mjs.json5_parser import dump_json5
from npm_mjs.json5_parser import encode_json5
from npm_mjs.json5_parser import load_json5
from npm_mjs.json5_parser import parse_json5


class TestJSON5ParserBasics(unittest.TestCase):
    """Test basic JSON5 parsing functionality."""

    def test_basic_unquoted_key(self):
        """Test parsing unquoted object keys."""
        content = '{ key: "value" }'
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value"})

    def test_quoted_key_with_special_chars(self):
        """Test parsing keys with special characters (already quoted)."""
        content = '{ "@rspack/core": "1.6.7" }'
        result = parse_json5(content)
        self.assertEqual(result, {"@rspack/core": "1.6.7"})

    def test_standard_json(self):
        """Test that standard JSON still works."""
        content = '{"name": "test", "version": "1.0.0"}'
        result = parse_json5(content)
        self.assertEqual(result, {"name": "test", "version": "1.0.0"})

    def test_nested_objects(self):
        """Test parsing nested objects."""
        content = """
        {
          dependencies: {
            "@rspack/core": "1.6.7",
            "other": "1.0.0"
          }
        }
        """
        result = parse_json5(content)
        self.assertEqual(
            result,
            {"dependencies": {"@rspack/core": "1.6.7", "other": "1.0.0"}},
        )

    def test_arrays(self):
        """Test parsing arrays."""
        content = '{ items: ["a", "b", "c"] }'
        result = parse_json5(content)
        self.assertEqual(result, {"items": ["a", "b", "c"]})

    def test_mixed_types(self):
        """Test parsing various JSON types."""
        content = """
        {
          string: "text",
          number: 42,
          float: 3.14,
          bool_true: true,
          bool_false: false,
          null_value: null
        }
        """
        result = parse_json5(content)
        self.assertEqual(
            result,
            {
                "string": "text",
                "number": 42,
                "float": 3.14,
                "bool_true": True,
                "bool_false": False,
                "null_value": None,
            },
        )


class TestJSON5Comments(unittest.TestCase):
    """Test comment removal functionality."""

    def test_single_line_comment_start(self):
        """Test single-line comment at the start."""
        content = """
        // This is a comment
        { key: "value" }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value"})

    def test_single_line_comment_inline(self):
        """Test single-line comment after a value."""
        content = """
        {
          key: "value", // This is a comment
          other: "value2"
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value", "other": "value2"})

    def test_multi_line_comment(self):
        """Test multi-line comment removal."""
        content = """
        /* This is a
           multi-line comment */
        { key: "value" }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value"})

    def test_multi_line_comment_inline(self):
        """Test inline multi-line comment."""
        content = """
        {
          /* comment */ key: "value"
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value"})

    def test_comment_with_url(self):
        """Test that URLs in comments don't break parsing."""
        content = """
        {
          // See https://example.com for more info
          name: "test",
          url: "https://github.com/user/repo" // GitHub repo
        }
        """
        result = parse_json5(content)
        self.assertEqual(
            result,
            {"name": "test", "url": "https://github.com/user/repo"},
        )

    def test_multi_line_comment_with_slashes(self):
        """Test multi-line comment containing // inside."""
        content = """
        {
          /* This is a comment with https://example.com
             and some // slashes too */
          name: "test"
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"name": "test"})


class TestJSON5StringsWithSlashes(unittest.TestCase):
    """Test that // inside strings is preserved correctly."""

    def test_double_slashes_in_string(self):
        """Test string with // that are not comments."""
        content = '{ path: "some//path//with//slashes" }'
        result = parse_json5(content)
        self.assertEqual(result, {"path": "some//path//with//slashes"})

    def test_url_in_string(self):
        """Test URL preservation in strings."""
        content = """
        {
          url: "https://example.com",
          homepage: "https://github.com/user/repo"
        }
        """
        result = parse_json5(content)
        self.assertEqual(
            result,
            {"url": "https://example.com", "homepage": "https://github.com/user/repo"},
        )

    def test_comment_marker_in_string(self):
        """Test that // in string is not treated as comment."""
        content = '{ comment: "This is not a // comment" }'
        result = parse_json5(content)
        self.assertEqual(result, {"comment": "This is not a // comment"})

    def test_mixed_slashes(self):
        """Test various slash patterns in strings."""
        content = """
        {
          single: "a/b/c",
          double: "a//b//c",
          url: "http://example.com",
          path: "C://Users//test"
        }
        """
        result = parse_json5(content)
        self.assertEqual(
            result,
            {
                "single": "a/b/c",
                "double": "a//b//c",
                "url": "http://example.com",
                "path": "C://Users//test",
            },
        )


class TestJSON5Quotes(unittest.TestCase):
    """Test single and double quote handling."""

    def test_single_quoted_string(self):
        """Test single-quoted string conversion."""
        content = "{ key: 'value' }"
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value"})

    def test_single_quotes_with_double_inside(self):
        """Test single-quoted string containing double quotes."""
        content = "{ single: 'value with \"quotes\"' }"
        result = parse_json5(content)
        self.assertEqual(result, {"single": 'value with "quotes"'})

    def test_double_quotes_with_single_inside(self):
        """Test double-quoted string containing single quotes."""
        content = "{ double: \"value with 'quotes'\" }"
        result = parse_json5(content)
        self.assertEqual(result, {"double": "value with 'quotes'"})

    def test_mixed_quote_types(self):
        """Test mixing single and double quotes."""
        content = """
        {
          single: 'value1',
          double: "value2",
          mixed1: 'has "double" inside',
          mixed2: "has 'single' inside"
        }
        """
        result = parse_json5(content)
        self.assertEqual(
            result,
            {
                "single": "value1",
                "double": "value2",
                "mixed1": 'has "double" inside',
                "mixed2": "has 'single' inside",
            },
        )

    def test_url_with_single_quotes(self):
        """Test URL in single-quoted string."""
        content = "{ url: 'https://example.com' }"
        result = parse_json5(content)
        self.assertEqual(result, {"url": "https://example.com"})


class TestJSON5EscapeSequences(unittest.TestCase):
    """Test escape sequence handling."""

    def test_escaped_double_quotes(self):
        """Test escaped double quotes in strings."""
        content = r'{ message: "He said \"hello\"" }'
        result = parse_json5(content)
        self.assertEqual(result, {"message": 'He said "hello"'})

    def test_escaped_backslash(self):
        """Test escaped backslashes."""
        content = r'{ path: "C:\\Users\\test\\path" }'
        result = parse_json5(content)
        self.assertEqual(result, {"path": r"C:\Users\test\path"})

    def test_escaped_newline(self):
        """Test escaped newline character."""
        content = r'{ text: "line1\nline2" }'
        result = parse_json5(content)
        self.assertEqual(result, {"text": "line1\nline2"})

    def test_escaped_tab(self):
        """Test escaped tab character."""
        content = r'{ text: "col1\tcol2" }'
        result = parse_json5(content)
        self.assertEqual(result, {"text": "col1\tcol2"})

    def test_mixed_escapes_and_slashes(self):
        """Test combination of escapes and slashes."""
        content = r'{ message: "He said \"hello // world\"" }'
        result = parse_json5(content)
        self.assertEqual(result, {"message": 'He said "hello // world"'})


class TestJSON5TrailingCommas(unittest.TestCase):
    """Test trailing comma handling."""

    def test_trailing_comma_object(self):
        """Test trailing comma in object."""
        content = """
        {
          key1: "value1",
          key2: "value2",
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"key1": "value1", "key2": "value2"})

    def test_trailing_comma_array(self):
        """Test trailing comma in array."""
        content = """
        {
          items: [
            "item1",
            "item2",
          ]
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"items": ["item1", "item2"]})

    def test_trailing_comma_nested(self):
        """Test trailing commas in nested structures."""
        content = """
        {
          obj: {
            key: "value",
          },
          arr: ["a", "b",],
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"obj": {"key": "value"}, "arr": ["a", "b"]})

    def test_multiple_trailing_commas(self):
        """Test multiple trailing commas in same structure."""
        content = """
        {
          a: "1",
          b: "2",
          c: "3",
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"a": "1", "b": "2", "c": "3"})


class TestJSON5ComplexScenarios(unittest.TestCase):
    """Test complex realistic scenarios."""

    def test_package_json5_example(self):
        """Test a realistic package.json5 file."""
        content = """
        // Package configuration
        {
          name: "my-package",
          version: "1.0.0",
          description: "A package with // in description",
          // Repository info
          repository: {
            type: "git",
            url: "https://github.com/user/repo.git" // Main repo
          },
          scripts: {
            test: "echo \\"test\\"",
            build: "npm run compile"
          },
          dependencies: {
            "@org/package": "^1.0.0",
            "another": "2.0.0", // Latest version
          },
          /* Multi-line comment
             with various content
             including https://example.com
             and // comment markers */
          author: "John Doe"
        }
        """
        result = parse_json5(content)
        self.assertEqual(result["name"], "my-package")
        self.assertEqual(result["version"], "1.0.0")
        self.assertEqual(result["description"], "A package with // in description")
        self.assertEqual(
            result["repository"]["url"],
            "https://github.com/user/repo.git",
        )
        self.assertEqual(result["scripts"]["test"], 'echo "test"')
        self.assertEqual(result["dependencies"]["@org/package"], "^1.0.0")
        self.assertEqual(result["author"], "John Doe")

    def test_npm_mjs_package_json5(self):
        """Test the actual package.json5 from this project."""
        content = """
        // Django-npm-mjs will combine this file with package.json files in other installed
        // apps before executing npm commands. Different from a regular package.json, comments
        // are allowed in this file.
        {
          description: "Install dependencies for ES6 transpilation",
          private: true,
          dependencies: {
            "@rspack/core": "1.6.7",
            "@rspack/cli": "1.6.7"
          },
        }
        """
        result = parse_json5(content)
        self.assertEqual(
            result["description"],
            "Install dependencies for ES6 transpilation",
        )
        self.assertTrue(result["private"])
        self.assertEqual(result["dependencies"]["@rspack/core"], "1.6.7")
        self.assertEqual(result["dependencies"]["@rspack/cli"], "1.6.7")

    def test_long_description_line(self):
        """Test parsing with a very long description (regression test for char 178 error)."""
        long_desc = (
            "This is a very long description that should push us past "
            "character 178 when combined with the other content"
        )
        content = f"""
        {{
          name: "test-package",
          version: "1.0.0",
          description: "{long_desc}",
          author: "Test Author",
          dependencies: {{
            "package1": "^1.0.0"
          }}
        }}
        """
        result = parse_json5(content)
        self.assertEqual(result["name"], "test-package")
        self.assertIn("very long description", result["description"])

    def test_inline_objects_and_arrays(self):
        """Test compact inline notation."""
        content = """
        {
          version: "1.0.0",
          engines: { node: ">=14.0.0", npm: ">=6.0.0" },
          keywords: ["django", "npm", "es6"],
          main: "index.js"
        }
        """
        result = parse_json5(content)
        self.assertEqual(result["engines"]["node"], ">=14.0.0")
        self.assertEqual(result["keywords"], ["django", "npm", "es6"])

    def test_empty_structures(self):
        """Test empty objects and arrays."""
        content = """
        {
          empty_obj: {},
          empty_arr: [],
          obj_with_empty: { nested: {} },
          arr_with_empty: [[]]
        }
        """
        result = parse_json5(content)
        self.assertEqual(result["empty_obj"], {})
        self.assertEqual(result["empty_arr"], [])
        self.assertEqual(result["obj_with_empty"]["nested"], {})
        self.assertEqual(result["arr_with_empty"], [[]])


class TestJSON5EdgeCases(unittest.TestCase):
    """Test edge cases and boundary conditions."""

    def test_comment_at_end_of_file(self):
        """Test comment at the very end."""
        content = """
        { key: "value" }
        // Final comment
        """
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value"})

    def test_multiple_keys_one_line(self):
        """Test multiple keys on a single line."""
        content = '{ key1: "value1", key2: "value2", key3: "value3" }'
        result = parse_json5(content)
        self.assertEqual(result, {"key1": "value1", "key2": "value2", "key3": "value3"})

    def test_whitespace_variations(self):
        """Test various whitespace patterns."""
        content = """
        {
          key1  :  "value1"  ,
          key2:"value2",
          key3 : "value3"
        }
        """
        result = parse_json5(content)
        self.assertEqual(result, {"key1": "value1", "key2": "value2", "key3": "value3"})

    def test_unicode_strings(self):
        """Test Unicode content in strings."""
        content = '{ text: "Hello 世界 🌍" }'
        result = parse_json5(content)
        self.assertEqual(result, {"text": "Hello 世界 🌍"})

    def test_numbers_with_underscores_not_supported(self):
        """Test that number separators are handled (pass through to JSON parser)."""
        # JSON5 supports underscores in numbers, but Python's json doesn't
        # Our parser just passes through, so this will fail in json.loads
        content = "{ number: 1_000_000 }"
        with self.assertRaises(json.JSONDecodeError):
            parse_json5(content)

    def test_glob_patterns(self):
        """Test glob patterns often found in package.json."""
        content = """
        {
          files: ["src/**/*.js", "dist/**/*.mjs"],
          pattern: "**/*.test.js"
        }
        """
        result = parse_json5(content)
        self.assertEqual(result["files"], ["src/**/*.js", "dist/**/*.mjs"])
        self.assertEqual(result["pattern"], "**/*.test.js")


class TestJSON5FileOperations(unittest.TestCase):
    """Test file loading and saving operations."""

    def test_load_json5_file(self):
        """Test loading a JSON5 file."""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json5", delete=False) as f:
            f.write('{ key: "value", number: 42 }')
            temp_path = f.name

        try:
            result = load_json5(temp_path)
            self.assertEqual(result, {"key": "value", "number": 42})
        finally:
            os.unlink(temp_path)

    def test_dump_json5_file(self):
        """Test dumping data to a JSON file."""
        data = {"key": "value", "number": 42}

        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            temp_path = f.name

        try:
            dump_json5(data, temp_path)

            # Read back and verify
            with open(temp_path) as f:
                content = f.read()
                result = json.loads(content)

            self.assertEqual(result, data)
        finally:
            os.unlink(temp_path)

    def test_encode_json5(self):
        """Test encoding data to JSON string."""
        data = {"key": "value", "nested": {"inner": "data"}}
        result = encode_json5(data)

        # Should be valid JSON
        parsed = json.loads(result)
        self.assertEqual(parsed, data)

        # Should be formatted with indentation
        self.assertIn("\n", result)
        self.assertIn("  ", result)


class TestJSON5ErrorHandling(unittest.TestCase):
    """Test error handling and edge cases."""

    def test_invalid_json_after_processing(self):
        """Test that invalid JSON raises appropriate error."""
        content = '{ key: "value", }'  # Trailing comma before }
        # This should actually work due to our trailing comma removal
        result = parse_json5(content)
        self.assertEqual(result, {"key": "value"})

    def test_unclosed_string(self):
        """Test that unclosed string raises error."""
        content = '{ key: "value }'
        with self.assertRaises(json.JSONDecodeError):
            parse_json5(content)

    def test_unclosed_object(self):
        """Test that unclosed object raises error."""
        content = '{ key: "value"'
        with self.assertRaises(json.JSONDecodeError):
            parse_json5(content)

    def test_unclosed_array(self):
        """Test that unclosed array raises error."""
        content = '{ arr: ["item1", "item2" }'
        with self.assertRaises(json.JSONDecodeError):
            parse_json5(content)

    def test_empty_string(self):
        """Test parsing empty string."""
        with self.assertRaises(json.JSONDecodeError):
            parse_json5("")

    def test_only_comment(self):
        """Test parsing file with only comments."""
        content = "// Just a comment"
        with self.assertRaises(json.JSONDecodeError):
            parse_json5(content)


if __name__ == "__main__":
    unittest.main()



import hashlib
import json
import os
import time
from subprocess import call

from django.apps import apps as django_apps
from django.core.management import call_command
from django.core.management.base import BaseCommand

from npm_mjs import signals
from npm_mjs.paths import SETTINGS_PATHS
from npm_mjs.paths import TRANSPILE_CACHE_PATH
from npm_mjs.tools import get_last_run
from npm_mjs.tools import set_last_run


def get_package_hash():
    """Generate a hash of all package.json files"""
    hash_md5 = hashlib.md5()
    for config in django_apps.get_app_configs():
        for filename in ["package.json", "package.json5"]:
            filepath = os.path.join(config.path, filename)
            if os.path.exists(filepath):
                with open(filepath, "rb") as f:
                    hash_md5.update(f.read())
    return hash_md5.hexdigest()


def install_npm(force, stdout, post_npm_signal=True):
    change_times = [0]
    for path in SETTINGS_PATHS:
        change_times.append(os.path.getmtime(path))
    settings_change = max(change_times)
    package_hash = get_package_hash()
    cache_file = os.path.join(TRANSPILE_CACHE_PATH, "package_hash.json")

    if os.path.exists(cache_file):
        with open(cache_file) as f:
            cached_hash = json.load(f).get("hash")
    else:
        cached_hash = None

    npm_install = False
    if (
        settings_change > get_last_run("npm_install")
        or package_hash != cached_hash
        or force
    ):
        stdout.write("Installing pnpm dependencies...")
        os.makedirs(TRANSPILE_CACHE_PATH, exist_ok=True)
        set_last_run("npm_install", int(round(time.time())))
        call_command("create_package_json")

        stdout.write("Installing dependencies...")
        call(["npx", "-y", "pnpm", "install"], cwd=TRANSPILE_CACHE_PATH)

        # Update cache
        with open(cache_file, "w") as f:
            json.dump({"hash": package_hash}, f)

        if post_npm_signal:
            signals.post_npm_install.send(sender=None)
        npm_install = True
    else:
        stdout.write("No changes detected, skipping pnpm install.")

    return npm_install


class Command(BaseCommand):
    help = "Run npm install on package.json files in app folders."

    def add_arguments(self, parser):
        parser.add_argument(
            "--force",
            action="store_true",
            dest="force",
            default=False,
            help="Force npm install even if no change is detected.",
        )

        parser.add_argument(
            "--nosignal",
            action="store_false",
            dest="post_npm_signal",
            default=True,
            help="Send a signal after finishing npm install.",
        )

    def handle(self, *args, **options):
        install_npm(options["force"], self.stdout, options["post_npm_signal"])



import json
import os
import shutil
import subprocess
import time
from subprocess import call
from urllib.parse import urljoin

from django.apps import apps
from django.conf import settings
from django.contrib.staticfiles import finders
from django.core.management.base import BaseCommand
from django.templatetags.static import PrefixNode

from .collectstatic import Command as CSCommand
from .npm_install import install_npm
from npm_mjs import signals
from npm_mjs.paths import PROJECT_PATH
from npm_mjs.paths import STATIC_ROOT
from npm_mjs.paths import TRANSPILE_CACHE_PATH
from npm_mjs.tools import set_last_run

# Run this script every time you update an *.mjs file or any of the
# modules it loads.

OLD_RSPACK_CONFIG_JS = ""

RSPACK_CONFIG_JS_PATH = os.path.join(TRANSPILE_CACHE_PATH, "rspack.config.js")

try:
    with open(RSPACK_CONFIG_JS_PATH) as file:
        OLD_RSPACK_CONFIG_JS = file.read()
except OSError:
    pass


class Command(BaseCommand):
    help = (
        "Transpile ES2015+ JavaScript to ES5 JavaScript + include NPM " "dependencies"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--force",
            action="store_true",
            dest="force",
            default=False,
            help="Force transpile even if no change is detected.",
        )

    def handle(self, *args, **options):
        if options["force"]:
            force = True
        else:
            force = False
        start = int(round(time.time()))
        npm_install = install_npm(force, self.stdout)
        js_paths = finders.find("js/", True)
        # Remove paths inside of collection dir
        js_paths = [x for x in js_paths if not x.startswith(STATIC_ROOT)]
        # Reverse list so that overrides function as expected. Static file from
        # first app mentioned in INSTALLED_APPS has preference.
        js_paths.reverse()

        transpile_path = os.path.join(PROJECT_PATH, "static-transpile")

        if os.path.exists(transpile_path):
            files = []
            for js_path in js_paths:
                for root, _dirnames, filenames in os.walk(js_path):
                    for filename in filenames:
                        files.append(os.path.join(root, filename))
            newest_file = max(files, key=os.path.getmtime)
            if (
                os.path.commonprefix([newest_file, transpile_path]) == transpile_path
                and not npm_install
                and not force
            ):
                # Transpile not needed as nothing has changed and not forced
                return
            # Remove any previously created static output dirs
            shutil.rmtree(transpile_path, ignore_errors=True)
        self.stdout.write("Transpiling...")
        os.makedirs(TRANSPILE_CACHE_PATH, exist_ok=True)
        # We reload the file as other values may have changed in the meantime
        set_last_run("transpile", start)
        # Create a static output dir
        out_dir = os.path.join(transpile_path, "js/")
        os.makedirs(out_dir, exist_ok=True)
        with open(os.path.join(transpile_path, "README.txt"), "w") as f:
            f.write(
                "These files have been automatically generated. "
                "DO NOT EDIT THEM! \n Changes will be overwritten. Edit "
                "the original files in one of the django apps, and run "
                "./manage.py transpile.",
            )

        mainfiles = []
        sourcefiles = []
        lib_sourcefiles = []
        for path in js_paths:
            for mainfile in (
                subprocess.check_output(
                    ["find", path, "-type", "f", "-name", "*.mjs", "-print"],
                )
                .decode("utf-8")
                .split("\n")[:-1]
            ):
                mainfiles.append(mainfile)
            for sourcefile in (
                subprocess.check_output(
                    ["find", path, "-type", "f", "-wholename", "*js"],
                )
                .decode("utf-8")
                .split("\n")[:-1]
            ):
                if "static/js" in sourcefile:
                    sourcefiles.append(sourcefile)
                if "static-libs/js" in sourcefile:
                    lib_sourcefiles.append(sourcefile)
        # Collect all JavaScript in a temporary dir (similar to
        # ./manage.py collectstatic).
        # This allows for the modules to import from oneanother, across Django
        # Apps.

        cache_path = os.path.join(TRANSPILE_CACHE_PATH, "js/")
        os.makedirs(cache_path, exist_ok=True)
        # Note all cache files so that we can remove outdated files that no
        # longer are in the prject.
        cache_files = []
        # Note all plugin dirs and the modules inside of them to crate index.js
        # files inside of them.
        plugin_dirs = {}
        for sourcefile in sourcefiles:
            relative_path = sourcefile.split("static/js/")[1]
            outfile = os.path.join(cache_path, relative_path)
            cache_files.append(outfile)
            dirname = os.path.dirname(outfile)
            os.makedirs(dirname, exist_ok=True)
            shutil.copyfile(sourcefile, outfile)
            # Check for plugin connectors
            if relative_path[:8] == "plugins/":
                if dirname not in plugin_dirs:
                    plugin_dirs[dirname] = []
                module_name = os.path.splitext(os.path.basename(relative_path))[0]
                if module_name != "init" and module_name not in plugin_dirs[dirname]:
                    plugin_dirs[dirname].append(module_name)

        for sourcefile in lib_sourcefiles:
            relative_path = sourcefile.split("static-libs/js/")[1]
            outfile = os.path.join(cache_path, relative_path)
            cache_files.append(outfile)
            dirname = os.path.dirname(outfile)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
                shutil.copyfile(sourcefile, outfile)
            elif not os.path.isfile(outfile):
                shutil.copyfile(sourcefile, outfile)
            elif os.path.getmtime(outfile) < os.path.getmtime(sourcefile):
                shutil.copyfile(sourcefile, outfile)

        # Write an index.js file for every plugin dir
        for plugin_dir in plugin_dirs:
            index_js = ""
            for module_name in plugin_dirs[plugin_dir]:
                index_js += 'export * from "./%s"\n' % module_name
            outfile = os.path.join(plugin_dir, "index.js")
            cache_files.append(outfile)
            if not os.path.isfile(outfile):
                index_file = open(outfile, "w")
                index_file.write(index_js)
                index_file.close()
            else:
                index_file = open(outfile)
                old_index_js = index_file.read()
                index_file.close()
                if old_index_js != index_js:
                    index_file = open(outfile, "w")
                    index_file.write(index_js)
                    index_file.close()

        # Check for outdated files that should be removed
        for existing_file in (
            subprocess.check_output(["find", cache_path, "-type", "f"])
            .decode("utf-8")
            .split("\n")[:-1]
        ):
            if existing_file not in cache_files:
                self.stdout.write("Removing %s" % existing_file)
                os.remove(existing_file)
        if apps.is_installed("django.contrib.staticfiles"):
            from django.contrib.staticfiles.storage import staticfiles_storage

            static_base_url = staticfiles_storage.base_url
        else:
            static_base_url = PrefixNode.handle_simple("STATIC_URL")
        transpile_base_url = urljoin(static_base_url, "js/")
        if (
            hasattr(settings, "RSPACK_CONFIG_TEMPLATE")
            and settings.RSPACK_CONFIG_TEMPLATE
        ):
            rspack_config_template_path = settings.RSPACK_CONFIG_TEMPLATE
        else:
            rspack_config_template_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "rspack.config.template.js",
            )
        entries = {}
        for mainfile in mainfiles:
            basename = os.path.basename(mainfile)
            modulename = basename.split(".")[0]
            file_path = os.path.join(cache_path, basename)
            entries[modulename] = file_path
        find_static = CSCommand()
        find_static.set_options(
            **{
                "interactive": False,
                "verbosity": 0,
                "link": False,
                "clear": False,
                "dry_run": True,
                "ignore_patterns": ["js/", "admin/"],
                "use_default_ignore_patterns": True,
                "post_process": True,
            },
        )
        found_files = find_static.collect()
        static_frontend_files = (
            found_files["modified"]
            + found_files["unmodified"]
            + found_files["post_processed"]
        )
        transpile = {
            "OUT_DIR": out_dir,
            "VERSION": start,
            "BASE_URL": transpile_base_url,
            "ENTRIES": entries,
            "STATIC_FRONTEND_FILES": [
                urljoin(static_base_url, x) for x in static_frontend_files
            ],
        }
        with open(rspack_config_template_path) as f:
            rspack_config_template = f.read()
        settings_dict = {}
        for var in dir(settings):
            if var in ["DATABASES", "SECRET_KEY"]:
                # For extra security, we do not copy DATABASES or SECRET_KEY
                continue
            try:
                settings_dict[var] = getattr(settings, var)
            except AttributeError:
                pass
        rspack_config_js = rspack_config_template.replace(
            "window.transpile",
            json.dumps(transpile),
        ).replace("window.settings", json.dumps(settings_dict, default=lambda x: False))

        if rspack_config_js is not OLD_RSPACK_CONFIG_JS:
            with open(RSPACK_CONFIG_JS_PATH, "w") as f:
                f.write(rspack_config_js)
        call(["./node_modules/.bin/rspack"], cwd=TRANSPILE_CACHE_PATH)
        end = int(round(time.time()))
        self.stdout.write("Time spent transpiling: " + str(end - start) + " seconds")
        signals.post_transpile.send(sender=None)



import pickle

from .paths import TRANSPILE_TIME_PATH

_last_run = {}


def load_last_name():
    global _last_run
    try:
        with open(TRANSPILE_TIME_PATH, "rb") as f:
            _last_run = {**_last_run, **pickle.load(f)}
    except (EOFError, OSError, TypeError):
        pass


def get_last_run(name):
    global _last_run
    if name not in _last_run:
        load_last_name()
        if name not in _last_run:
            return 0
    return _last_run[name]


def set_last_run(name, timestamp):
    global _last_run
    load_last_name()
    _last_run[name] = timestamp
    with open(TRANSPILE_TIME_PATH, "wb") as f:
        pickle.dump(_last_run, f)



from base.management import BaseCommand
from django.core.management.commands import makemessages

# This makes makemessages create both translations for Python and JavaScript
# code in one go.

# Note that translating JavaScript files with template strings requires xgettext 0.24 (see README.md))


class Command(makemessages.Command, BaseCommand):

    def add_arguments(self, parser):
        # Call the parent class's add_arguments first
        super().add_arguments(parser)

        # Modify the domain argument's help text
        for action in parser._actions:
            if action.dest == "domain":
                action.help = (
                    "Domain of the messages file ('django' or 'djangojs'). "
                    "By default, both domains will be processed."
                )
                action.default = None  # This indicates we'll handle both domains
            elif action.dest == "locale":
                action.help = "Locale(s) to process. If none are specified, all locales will be processed."
            elif action.dest == "all":
                action.help = (
                    "Process all locales. "
                    "If not specified, and no locales are provided, all locales will be processed."
                )
                action.default = False

    def handle(self, *args, **options):
        options["ignore_patterns"] += [
            "venv",
            ".direnv",
            "node_modules",
            "static-transpile",
        ]
        if len(options["locale"]) == 0 and not options["all"]:
            options["all"] = True
        if options["domain"]:
            self.stdout.write("Domain %s" % options["domain"])
            return super().handle(*args, **options)
        else:
            options["domain"] = "django"
            self.stdout.write("Domain %s" % options["domain"])
            super().handle(*args, **options)
            options["domain"] = "djangojs"
            self.stdout.write("Domain %s" % options["domain"])
            super().handle(*args, **options)



import os

from setuptools import find_packages
from setuptools import setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(__file__, os.pardir)))

setup(
    packages=find_packages(),
    include_package_data=True,
)
